package org.xbib.interlibrary.catalog.builder;

import java.io.Closeable;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.interlibrary.api.action.worker.Tool;
import org.xbib.settings.Settings;

public abstract class Converter<R, P extends ConverterWorkerPool<R>>
        implements Tool, Closeable {

    private static final Logger logger = Logger.getLogger(Converter.class.getName());

    private P workerPool;

    private final AtomicBoolean closed;

    private volatile int returncode;

    public Converter() {
        closed = new AtomicBoolean();
    }

    @Override
    public void run(Settings settings) {
        try {
            returncode = 0;
            logger.log(Level.FINE, () -> "settings = " + settings.getAsMap());
            this.workerPool = newPool(settings);
            logger.log(Level.FINE, () -> "opening output");
            workerPool.openOutput();
            logger.log(Level.FINE, () -> "opening worker pool");
            workerPool.open();
            logger.log(Level.FINE, () -> "opening input");
            workerPool.openInput();
            logger.log(Level.FINE, () -> "opening metrics");
            workerPool.openMetrics();
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            returncode = 1;
        } finally {
            try {
                close();
            } catch (IOException ioe) {
                logger.log(Level.SEVERE, ioe.getMessage(), ioe);
            }
        }
    }

    @Override
    public int getReturnCode() {
        return returncode;
    }

    @Override
    public void close() throws IOException {
        if (closed.compareAndSet(false, true)) {
            try {
                if (workerPool != null) {
                    logger.log(Level.INFO, () -> "closing worker pool, return code = " + returncode);
                    workerPool.setReturnCode(returncode);
                    workerPool.close();
                    for (Map.Entry<Runnable, Throwable> entry : workerPool.getExceptions().entrySet()) {
                        Runnable r = entry.getKey();
                        Throwable t = entry.getValue();
                        logger.log(Level.SEVERE, MessageFormat.format("{0}: {1}", r, t.getMessage()), t);
                    }
                    if (!workerPool.getExceptions().isEmpty()) {
                        throw new IOException("there were worker errors");
                    }
                }
            } catch (Throwable t) {
                logger.log(Level.SEVERE, t.getMessage(), t);
                returncode = 1;
            }
        }
    }

    public abstract P newPool(Settings settings) throws IOException;

}
