package org.xbib.interlibrary.catalog.builder;

import java.io.BufferedInputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.UnknownServiceException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.zip.GZIPInputStream;
import org.xbib.settings.Settings;

public class ConverterInput<R> implements Closeable {

    private static final Logger logger = Logger.getLogger(ConverterInput.class.getName());

    private static final int DEFAULT_BUFFER_SIZE = 1024 * 1024;

    private final Map<String, List<String>> fileMap = new HashMap<>();

    public ConverterInput() {
    }

    public Map<String, List<String>> getFileMap() {
        return fileMap;
    }

    public void createInputFromFileMap(Settings settings, ConverterWorkerPool<R> pool) throws IOException {
        Map<String, Settings> inputMap = settings.getGroups("input");
        for (Map.Entry<String, Settings> entry : inputMap.entrySet()) {
            String inputKey = entry.getKey();
            Settings inputSettings = entry.getValue();
            if (inputSettings.getAsArray("uri").length > 0) {
                String[] inputs = inputSettings.getAsArray("uri");
                logger.log(Level.INFO, () -> MessageFormat.format("{0}: preparing {1} inputs from {2}",
                        inputKey, inputs.length, Arrays.asList(inputs)));
                for (String input : inputs) {
                    if (inputKey.startsWith("queue")) {
                        pool.submit(pool.newWorkerRequest(input, inputKey));
                    } else {
                        logger.log(Level.WARNING, () -> "not putting into queue because key '" + inputKey + "' does not start with 'queue'");
                        fileMap.put(inputKey, Arrays.asList(inputs));
                    }
                }
            } else if (inputSettings.get("uri") != null) {
                String input = settings.get("uri");
                boolean parallel = inputSettings.getAsBoolean("parallel", false);
                int worker = inputSettings.getAsInt("worker", 1);
                logger.log(Level.INFO, () -> MessageFormat.format("{0}: preparing one input from {1} (parallel={2},worker={3})",
                        inputKey, input, parallel, worker));
                if (inputKey.startsWith("queue")) {
                    pool.submit(pool.newWorkerRequest(input, inputKey));
                    if (parallel) {
                        for (int i = 1; i < worker; i++) {
                            pool.submit(pool.newWorkerRequest(input, inputKey));
                        }
                    }
                } else {
                    String fileMapKey = inputSettings.get("id") != null ? inputSettings.get("id") : inputKey;
                    logger.log(Level.WARNING, () -> "not putting into queue because key '" + fileMapKey + "' does not start with 'queue'");
                    fileMap.put(fileMapKey, Collections.singletonList(input));
                }
            } else if (inputSettings.get("pattern") != null) {
                String base = inputSettings.get("base");
                String basepattern = inputSettings.get("basepattern");
                String path = inputSettings.get("path");
                String pathpattern = inputSettings.get("pattern");
                String since = inputSettings.get("since");
                // default: queue all files, ordered by last modified timestamp in ascending order
                Queue<String> uris = new Finder()
                        .find(base, basepattern, path, pathpattern, since)
                        .sortBy(inputSettings.get("sort_by", "lastmodified"))
                        .order(inputSettings.get("order", "asc"))
                        .getURIs(inputSettings.getAsInt("max", -1));
                List<String> top10 = uris.stream().limit(10).toList();
                logger.log(Level.INFO, () -> MessageFormat.format("number of URIs found = {0}, top ten = {1} ",
                        uris.size(), top10));
                if (inputKey.startsWith("queue")) {
                    for (String uri : uris) {
                        pool.submit(pool.newWorkerRequest(uri, uri));
                    }
                } else {
                    logger.log(Level.WARNING, () -> "not putting into queue because key '" + inputKey + "' does not start with 'queue'");
                    fileMap.put(inputKey, new ArrayList<>(uris));
                }
            } else if (inputSettings.get("archive") != null) {
                String archive = inputSettings.get("archive");
                logger.log(Level.INFO, () -> MessageFormat.format("{0}: preparing inputs from archive={1}",
                        inputKey, archive));
                if (inputKey.startsWith("queue")) {
                    pool.submit(pool.newWorkerRequest(archive, archive));
                } else {
                    logger.log(Level.WARNING, () -> "not putting into queue because key '" + inputKey + "' does not start with 'queue'");
                    fileMap.put(inputKey, Collections.singletonList(archive));
                }
            } else {
                String fileMapKey = inputSettings.get("id") != null ? inputSettings.get("id") : inputKey;
                List<String> l = inputSettings.get("uri") != null ?
                        Arrays.asList(inputSettings.getAsArray("uri")) :
                        Arrays.stream(inputSettings.getAsArray("name")).map(s -> "file://" + s).collect(Collectors.toList());
                fileMap.put(fileMapKey, l);
            }
        }
    }

    public static InputStream getInputStream(String uriString, Settings settings) throws IOException {
        if (uriString == null) {
            return null;
        }
        int buffersize = settings.getAsInt("buffersize", DEFAULT_BUFFER_SIZE);
        InputStream in;
        URI uri = null;
        try {
            uri = URI.create(uriString);
            if (!uri.isAbsolute()) {
                uri = URI.create("file://" + uriString);
            }
            in = new BufferedInputStream(uri.toURL().openStream(), buffersize);
        } catch (MalformedURLException | UnknownServiceException e) {
            logger.log(Level.FINE, e.getMessage(), e);
            // ordinary file path?
            Path path = Paths.get(uri.getSchemeSpecificPart());
            if (path.toFile().isFile()) {
                in = new BufferedInputStream(Files.newInputStream(path), buffersize);
            } else {
                throw new IOException("can't open for input, check file existence or access rights: "
                        + uri.getSchemeSpecificPart());
            }
        }
        if (uri.getSchemeSpecificPart().endsWith(".gz")) {
            in = new GZIPInputStream(in, buffersize);
        }
        return in;
    }

    @Override
    public void close() throws IOException {
        // nothing to do
    }
}
