package org.xbib.interlibrary.catalog.builder;

import java.io.IOException;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.text.MessageFormat;
import java.util.LinkedHashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.api.TimeValue;
import org.xbib.interlibrary.catalog.builder.util.FormatUtil;
import org.xbib.metrics.common.Meter;
import org.xbib.settings.Settings;

public class ConverterMetrics {

    private static final Logger logger = Logger.getLogger(ConverterMetrics.class.getName());

    protected final Settings settings;

    protected final ScheduledExecutorService service;

    protected final Map<String, MetricWriter> writers;

    protected Meter meter;

    protected TimeValue interval;

    protected String type;

    public ConverterMetrics(Settings settings) {
       this(settings, 4);
    }

    public ConverterMetrics(Settings settings, int corePoolSize) {
        this.settings = settings;
        this.service = Executors.newScheduledThreadPool(corePoolSize);
        this.writers = new LinkedHashMap<>();
    }

    public void openMetrics() {
        Objects.requireNonNull(settings, "settings must not be null");
        this.type = settings.get("metrics.type", "converter");
        this.interval = settings.getAsTime("metrics.interval", TimeValue.timeValueSeconds(10L));
        for (Map.Entry<String, Settings> entry : settings.getGroups("metrics").entrySet()) {
            try {
                String type = entry.getKey();
                Settings metricSettings = entry.getValue();
                String name = metricSettings.get("name", entry.getKey());
                Path path = Paths.get(name);
                MetricWriter metricWriter = new MetricWriter();
                metricWriter.setType(type);
                metricWriter.setPath(path);
                metricWriter.setWriter(Files.newBufferedWriter(path, StandardOpenOption.CREATE, StandardOpenOption.APPEND));
                metricWriter.setSettings(entry.getValue());
                metricWriter.setChart(Paths.get(metricSettings.get("chart", name + ".svg")));
                metricWriter.setTitle(metricSettings.get("title", type));
                if (metricSettings.containsSetting("locale")) {
                    metricWriter.setLocale(Locale.of(metricSettings.get("locale")));
                } else {
                    metricWriter.setLocale(Locale.getDefault());
                }
                if (!writers.containsKey(name)) {
                    writers.put(name, metricWriter);
                }
            } catch (IOException e) {
                logger.log(Level.WARNING, e.getMessage(), e);
            }
        }
        if (!writers.isEmpty()) {
            logger.log(Level.INFO, () -> "metrics prepared: " + writers.keySet());
        }
        Meter meter = new Meter(service);
        meter.start(interval.getSeconds());
        this.meter = meter;
        service.scheduleAtFixedRate(() -> append(meter), 0L, interval.getSeconds(), TimeUnit.SECONDS);
        logger.log(Level.INFO, () -> "scheduled metrics at interval of: " + interval);
    }

    public Meter getMeter() {
        return meter;
    }

    public void mark() {
        if (meter != null) {
            meter.mark();
        }
    }

    public void closeMetrics() {
        if (meter != null) {
            meter.shutdown();
        }
        service.shutdownNow();
        for (Map.Entry<String, MetricWriter> entry : writers.entrySet()) {
            try {
                MetricWriter metricWriter = entry.getValue();
                Writer writer = metricWriter.getWriter();
                if (writer != null) {
                    writer.close();
                    logger.log(Level.INFO, () -> entry.getKey() + " closed");
                }
            } catch (IOException e) {
                logger.log(Level.WARNING, e.getMessage(), e);
            }
        }
    }

    protected void append(Meter meter) {
        long docs = meter.getCount();
        long elapsed = meter.elapsed() / 1000000; // nanos to millis
        double dps = docs * 1000.0 / elapsed;
        long mean = Math.round(meter.getMeanRate());
        long oneminute = Math.round(meter.getOneMinuteRate());
        long fiveminute = Math.round(meter.getFiveMinuteRate());
        long fifteenminute = Math.round(meter.getFifteenMinuteRate());
        logger.log(Level.INFO, () -> MessageFormat.format("{0}: {1} docs, {2} ms = {3}, {4} = {5}, {6} ({7} {8} {9})",
                type,
                docs,
                elapsed,
                FormatUtil.formatDurationWords(elapsed, true, true),
                dps,
                FormatUtil.formatDocumentSpeed(dps),
                mean,
                oneminute,
                fiveminute,
                fifteenminute
        ));
        for (Map.Entry<String, MetricWriter> entry : writers.entrySet()) {
            try {
                MetricWriter metricWriter = entry.getValue();
                if (type != null && type.equals(metricWriter.getType()) && metricWriter.getWriter() != null) {
                    Settings settings = metricWriter.getSettings();
                    Locale locale = metricWriter.getLocale();
                    String format = settings.get("format", "%s\t%d\t%d\n");
                    String message = String.format(locale, format, metricWriter.getTitle(), elapsed, docs);
                    metricWriter.getWriter().write(message);
                    metricWriter.getWriter().flush();
                }
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }
    }
}
