package org.xbib.interlibrary.catalog.builder;

import java.io.BufferedOutputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.NoSuchFileException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.nio.file.StandardOpenOption;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.Map;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.Deflater;
import java.util.zip.GZIPOutputStream;
import org.xbib.settings.Settings;

public class ConverterOutput implements Closeable {

    private static final Logger logger = Logger.getLogger(ConverterOutput.class.getName());

    private final Map<String, Entry> fileMap = new HashMap<>();

    public ConverterOutput() {
    }

    public Map<String, Entry> getMap() {
        return fileMap;
    }

    /**
     * Structure of output entry:
     *
     * output:
     *   file:
     *     name: test
     *     path: test.xml
     *
     * @param outputSettings the output settings
     * @throws IOException if the output can not be created
     */
    public void createOutputFromFileMap(Settings outputSettings) throws IOException {
        Map<String, Settings> output = outputSettings.getGroups("output");
        for (Map.Entry<String, Settings> entry : output.entrySet()) {
            String key = entry.getKey();
            Settings settings = entry.getValue();
            if (!"file".equals(key)) {
                continue;
            }
            String name = settings.get("name");
            if (name == null) {
                logger.log(Level.WARNING, () -> "no name for output given, skipping");
                continue;
            }
            String pathName = settings.get("path", name);
            // use a tmp suffix
            Path path = Paths.get(pathName + ".tmp");
            boolean overwrite = settings.getAsBoolean("overwrite", true);
            // default must be false since overwrite default is true, this overrides overwrite
            boolean append = settings.getAsBoolean("append", false);
            StandardOpenOption option1 = overwrite ?
                    StandardOpenOption.CREATE : StandardOpenOption.CREATE_NEW;
            StandardOpenOption option2 = append ?
                    StandardOpenOption.APPEND : StandardOpenOption.TRUNCATE_EXISTING;
            OutputStream outputStream = Files.newOutputStream(path, option1, option2);
            if (pathName.endsWith(".gz")) {
                outputStream = new CompressedOutputStream(outputStream, settings.getAsInt("buffersize", 65536));
            } else {
                outputStream = new BufferedOutputStream(outputStream);
            }
            logger.log(Level.INFO, () -> "opening for write: " + name);
            fileMap.put(name, new Entry(name, settings, path, outputStream));
        }
    }

    public void closeFileMap(int returncode) throws IOException {
        for (Map.Entry<String, Entry> entry : fileMap.entrySet()) {
            Entry e = entry.getValue();
            logger.log(Level.INFO, MessageFormat.format("closing writes to {0} on path {1}",
                    entry.getKey(), e.getPath()));
            e.getOut().close();
            // important, perform renaming and retention policy only if success so far
            if (returncode == 0) {
                // rename from .tmp
                String fileName = e.getSettings().get("path", e.getSettings().get("name", e.getName()));
                Path dest = Paths.get(fileName);
                if (!e.getPath().toString().equals(dest.toString()) && !e.getPath().toString().contains("%")) {
                    logger.log(Level.INFO, () -> MessageFormat.format("renaming path from {0} to {1}",
                            e.getPath().toString(), dest.toString()));
                    if (Files.exists(e.getPath())) {
                        // allow overwriting old file but do not fail on error
                        try {
                            Files.move(e.getPath(), dest, StandardCopyOption.REPLACE_EXISTING);
                        } catch (IOException ee) {
                            logger.log(Level.WARNING, "unable to rename to " + dest);
                        }
                    }
                }
                if (e.getSettings().getAsBoolean("retention.enabled", false)) {
                    logger.log(Level.INFO, "performing retention policy");
                    performRetention(e);
                } else {
                    logger.log(Level.INFO, "retention disabled by configuration");
                }
            }
        }
    }

    private void performRetention(Entry entry) throws IOException {
        Settings settings = entry.getSettings().getAsSettings("retention");
        Path parent = entry.getPath().getParent();
        String basepattern = settings.get("basepattern", "*");
        String pattern = settings.get("pattern", "*");
        long minkeep = settings.getAsLong("minkeep", 1L);
        logger.log(Level.INFO, () -> MessageFormat.format("found retention policy: parent={0} basepattern={1} pattern={2} minkeep={3}",
                parent, pattern, minkeep));
        // sort by last modified date in descending order and skip min-to-keep entries
        Queue<Finder.PathFile> paths = new Finder()
                .find(parent, basepattern, entry.getPath(), pattern)
                .sortBy(settings.get("sort_by", "lastmodified"))
                .order(settings.get("order", "desc"))
                .skipPathFiles(minkeep);
        logger.log(Level.INFO, () -> MessageFormat.format("files to delete due to retention policy: {0}", paths));
        for (Finder.PathFile pf : paths) {
            try {
                Files.deleteIfExists(pf.getPath());
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }
    }

    @Override
    public void close() throws IOException {
        closeFileMap(0);
    }

    /**
     *
     */
    public static class Entry {

        private final String name;

        private final Settings settings;

        private final Path path;

        private final OutputStream out;

        Entry(String name, Settings settings, Path path, OutputStream out) {
            this.name = name;
            this.settings = settings;
            this.path = path;
            this.out = out;
        }

        public String getName() {
            return name;
        }

        public Settings getSettings() {
            return settings;
        }

        public Path getPath() {
            return path;
        }

        public OutputStream getOut() {
            return out;
        }

        public void copy(InputStream inputStream) throws IOException {
            inputStream.transferTo(out);
        }
    }

    /**
     * A GZIP output stream, modified for best compression.
     */
    private static class CompressedOutputStream extends GZIPOutputStream {

        CompressedOutputStream(OutputStream out, int size) throws IOException {
            super(out, size, true);
            def.setLevel(Deflater.BEST_COMPRESSION);
        }
    }
}
