package org.xbib.interlibrary.catalog.builder;

import org.xbib.time.chronic.Chronic;
import org.xbib.time.chronic.Span;

import java.io.IOException;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.FileVisitOption;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.PathMatcher;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.FileTime;
import java.text.ParseException;
import java.time.Instant;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * A finder for traversing a NIO file system.
 */
public class Finder {

    private static final Logger logger = Logger.getLogger(Finder.class.getName());

    private final FileSystem fileSystem;

    private final List<PathFile> input;

    private Comparator<PathFile> comparator;

    private int depth = Integer.MAX_VALUE;

    public Finder() {
        this(FileSystems.getDefault());
    }

    public Finder(FileSystem fileSystem) {
        this.fileSystem = fileSystem;
        this.input = new ArrayList<>();
    }

    public Finder find(String path, String pattern) throws IOException {
        return find(null, null, fileSystem.getPath(path), pattern);
    }

    public Finder find(String base, String basePattern, String path, String pathPattern) throws IOException {
        return find(base == null || base.isEmpty() ? null : fileSystem.getPath(base), basePattern,
                path == null || path.isEmpty() ? null : fileSystem.getPath(path), pathPattern);
    }

    public Finder find(Path base, String basePattern, Path path, String pathPattern) throws IOException {
        return find(base, basePattern, path, pathPattern, null);
    }

    public Finder find(String base, String basePattern, String path, String pathPattern, String since) throws IOException {
        return find(base == null || base.isEmpty() ? null : fileSystem.getPath(base), basePattern,
                path == null || path.isEmpty() ? null : fileSystem.getPath(path), pathPattern,
                since != null ? FileTime.from(fromChrono(since)) : null);
    }

    /**
     * Find the most recent version of a file/archive.
     *
     * @param base        the path of the base directory
     * @param basePattern a pattern to match directory entries in the base directory or null to match '*'
     * @param path        the path of the file/archive if no recent path can be found in the base directory
     * @param pathPattern     the file name pattern to match
     * @param modifiedSince match only if modified since this FileTime or null if check should skipped
     * @return this Finder
     * @throws IOException if find fails
     */
    public Finder find(Path base, String basePattern, Path path, String pathPattern, FileTime modifiedSince)
            throws IOException {
        if (base != null && path == null) {
            List<PathFile> baseMatches = new ArrayList<>();
            final PathMatcher baseMatcher = base.getFileSystem()
                    .getPathMatcher("glob:" + (basePattern != null ? basePattern : "*"));
            final PathMatcher pathMatcher = base.getFileSystem()
                    .getPathMatcher("glob:" + (pathPattern != null ? pathPattern : "*"));
            List<Path> paths = new ArrayList<>();
            try (Stream<Path> stream = Files.find(base, depth, (p, a) -> {
                        if (Files.isDirectory(p) && baseMatcher.matches(p.getFileName())) {
                            baseMatches.add(new PathFile(p, a));
                        }
                        boolean b = Files.isRegularFile(p) && pathMatcher.matches(p.getFileName());
                        if (b) {
                            input.add(new PathFile(p, a));
                        }
                        return b;
                    },
                    FileVisitOption.FOLLOW_LINKS)) {
                paths.addAll(stream.toList());
            }
            if (baseMatches.isEmpty()) {
                return this;
            }
            Comparator<PathFile> c = Comparator.comparing(o -> o.getAttributes().lastModifiedTime());
            baseMatches.sort(c.reversed());
            path = baseMatches.getFirst().getPath();
            final Path p = path;
            logger.log(Level.INFO, () -> "found most recent path: " + p);
        }
        if (path == null) {
            logger.log(Level.WARNING, "no path given");
            return this;
        }
        PathMatcher pathMatcher = path.getFileSystem()
                .getPathMatcher("glob:" + (pathPattern != null ? pathPattern : "*"));
        if (!Files.exists(path)) {
            return this;
        }
        try (Stream<Path> stream = walk(path,0)) {
            stream.forEach(p -> {
                try {
                    BasicFileAttributes attributes = Files.readAttributes(p, BasicFileAttributes.class);
                    if ((Files.isRegularFile(p) && pathMatcher.matches(p.getFileName())) &&
                            (modifiedSince == null || attributes.lastModifiedTime().toMillis() > modifiedSince.toMillis())) {
                        input.add(new PathFile(p, attributes));
                    }
                } catch (IOException e) {
                    logger.log(Level.WARNING, "unable to get file attributes for " + p);
                }
            });
        }
        return this;
    }

    public Finder depth(int depth) {
        this.depth = depth;
        return this;
    }

    public Finder sortBy(String mode) {
        if ("lastmodified".equals(mode)) {
            this.comparator = Comparator.comparing(p -> p.getAttributes().lastModifiedTime());
        } else if ("name".equals(mode)) {
            this.comparator = Comparator.comparing(p -> p.getPath().toString());
        }
        return this;
    }

    public Finder order(String mode) {
        if ("desc".equals(mode)) {
            this.comparator = Collections.reverseOrder(comparator);
        }
        return this;
    }

    public Queue<PathFile> getPathFiles() {
        return getPathFiles(-1);
    }

    public Queue<PathFile> getPathFiles(long max) {
        if (comparator != null) {
            input.sort(comparator);
        }
        return input.stream()
                .limit(max < 0 ? input.size() : max)
                .collect(Collectors.toCollection(ConcurrentLinkedQueue::new));
    }

    public Queue<PathFile> skipPathFiles(long skip) {
        if (comparator != null) {
            input.sort(comparator);
        }
        return input.stream()
                .skip(skip < 0 ? 0 : skip)
                .collect(Collectors.toCollection(ConcurrentLinkedQueue::new));
    }

    public Queue<String> getURIs() {
        return getURIs(-1);
    }

    public Queue<String> getURIs(long max) {
        if (comparator != null) {
            input.sort(comparator);
        }
        return input.stream()
                .map(p -> p.getPath().toAbsolutePath().toString())
                .limit(max < 0 ? input.size() : max)
                .collect(Collectors.toCollection(ConcurrentLinkedQueue::new));
    }

    private Stream<Path> walk(Path p, int depth) {
        if (depth > this.depth) {
            return Stream.empty();
        } else {
            if (Files.isReadable(p)) {
                if (Files.isDirectory(p)) {
                    try {
                        return Files.list(p).flatMap(q -> walk(q, depth + 1));
                    } catch (IOException e) {
                        logger.log(Level.FINE, "skipped " + p + " because of " + e.getMessage());
                        return Stream.empty();
                    }
                } else {
                    return Stream.of(p);
                }
            } else {
                logger.log(Level.WARNING, "unreadable: " + p);
            }
            return Stream.empty();
        }
    }

    private Instant fromChrono(String chronoString) {
        if (chronoString == null) {
            logger.log(Level.WARNING, "chrono string is null, assuming current time");
            return Instant.now();
        }
        try {
            Span span = Chronic.parse(chronoString);
            logger.log(Level.INFO, "found instant from chrono string: " +
                    DateTimeFormatter.ISO_INSTANT.format(span.getBeginCalendar()));
            return span.getBeginCalendar().toInstant();
        } catch (ParseException e) {
            logger.log(Level.WARNING, "no chrono string found in " + chronoString + ", parsing instant");
            return Instant.parse(chronoString);
        }
    }

    public static class PathFile {
        private final Path path;
        private final BasicFileAttributes attr;

        public PathFile(Path path, BasicFileAttributes attr) {
            this.path = path;
            this.attr = attr;
        }

        public Path getPath() {
            return path;
        }

        public BasicFileAttributes getAttributes() {
            return attr;
        }

        @Override
        public String toString() {
            return path.toString();
        }
    }
}
