package org.xbib.interlibrary.catalog.builder;

import java.io.BufferedReader;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.io.UncheckedIOException;
import java.nio.charset.StandardCharsets;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcListener;
import org.xbib.marc.label.RecordLabel;
import org.xbib.marc.transformer.field.MarcFieldTransformers;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xml.sax.InputSource;

/**
 * Read MARC from XML lines format.
 */
public class MarcXchangeXMLLinesReader implements Closeable {

    private static final Logger logger = Logger.getLogger(MarcXchangeXMLLinesReader.class.getName());

    private static final int DEFAULT_BUFFER_SIZE = 65536;

    private final Reader reader;

    private final MarcListener listener;

    private final int bufferSize;

    private MarcFieldTransformers marcFieldTransformers;

    private MarcValueTransformers marcValueTransformers;

    public MarcXchangeXMLLinesReader(InputStream inputStream) {
        this(inputStream, null);
    }

    public MarcXchangeXMLLinesReader(InputStream inputStream, MarcListener listener) {
        this(inputStream, DEFAULT_BUFFER_SIZE, listener);
    }

    public MarcXchangeXMLLinesReader(InputStream inputStream, int buffersize, MarcListener listener) {
        this(new InputStreamReader(inputStream, StandardCharsets.UTF_8), buffersize, listener);
    }

    public MarcXchangeXMLLinesReader(Reader reader, MarcListener listener) {
        this(reader, DEFAULT_BUFFER_SIZE, listener);
    }

    public MarcXchangeXMLLinesReader(Reader reader, int bufferSize, MarcListener listener) {
        this.reader = reader;
        this.bufferSize = bufferSize;
        this.listener = listener == null ? new EmptyListener() : listener;
        logger.log(Level.INFO, "listener = " + this.listener);
    }

    public void setMarcFieldTransformers(MarcFieldTransformers marcFieldTransformers) {
        this.marcFieldTransformers = marcFieldTransformers;
    }

    public void setMarcValueTransformers(MarcValueTransformers marcValueTransformers) {
        this.marcValueTransformers = marcValueTransformers;
    }

    public void parse() throws IOException {
        try (BufferedReader bufferedReader = new BufferedReader(reader, bufferSize)) {
            listener.beginCollection();
            bufferedReader.lines().forEach(this::parse);
            listener.endCollection();
        }
    }

    protected void parse(String line) {
        try {
            InputSource inputSource = new InputSource(new StringReader(line));
            Marc.builder()
                    .setMarcFieldTransformers(marcFieldTransformers)
                    .setMarcValueTransformers(marcValueTransformers)
                    .setMarcListener(listener)
                    .build()
                    .xmlReader().parse(inputSource);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    @Override
    public void close() throws IOException {
        reader.close();
    }

    private static class EmptyListener implements MarcListener {

        @Override
        public void beginCollection() {
            // do nothing
        }

        @Override
        public void beginRecord(String format, String type) {
            // do nothing
        }

        @Override
        public void leader(RecordLabel label) {
            // do nothing
        }

        @Override
        public void field(MarcField field) {
            // do nothing
        }

        @Override
        public void endRecord() {
            // do nothing
        }

        @Override
        public void endCollection() {
            // do nothing
        }
    }
}
