package org.xbib.interlibrary.catalog.builder.alma;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardCopyOption;
import java.time.Instant;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContent;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.CatalogEntityRdfOutput;
import org.xbib.interlibrary.catalog.builder.Converter;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.ConverterOutput;
import org.xbib.interlibrary.catalog.builder.ConverterWorker;
import org.xbib.interlibrary.catalog.builder.ConverterWorkerPool;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.builder.marc.DefaultWorkerPoolMarcRecordListener;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.io.archive.tar.TarArchiveEntry;
import org.xbib.io.archive.tar.TarArchiveInputStream;
import org.xbib.marc.Marc;
import org.xbib.marc.xml.MarcContentHandler;
import org.xbib.settings.Settings;

public class HbzAlma extends Converter<String, HbzAlma.HbzAlmaWorkerPool> {

    public HbzAlma() {
    }

    @Override
    public HbzAlmaWorkerPool newPool(Settings settings) throws IOException {
        return new HbzAlmaWorkerPool(settings);
    }

    public class HbzAlmaWorkerPool extends ConverterWorkerPool<String> {

        private final Logger logger = Logger.getLogger(HbzAlmaWorkerPool.class.getName());

        CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool;

        public HbzAlmaWorkerPool(Settings settings) throws IOException {
            super("hbz-alma", settings);
            catalogEntityWorkerPool = new CatalogEntityWorkerPool<>(settings,
                    new DefaultWorkerPoolMarcRecordListener(),
                    HbzAlmaCatalogEntityBuilder::new,
                    this);
        }

        @Override
        public HbzAlmaWorker newWorker(String name, int count) {
            return new HbzAlmaWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        RouteRdfXContent.RouteHandler newHandler() {
            return (content, p) -> {
                for (Map.Entry<String, ConverterOutput.Entry> entry : getConverterOutput().getMap().entrySet()) {
                    if (entry.getKey().startsWith("test")) {
                        try {
                            synchronized (this) {
                                entry.getValue().getOut().write(content.getBytes(StandardCharsets.UTF_8));
                                entry.getValue().getOut().write('\n');
                                entry.getValue().getOut().flush();
                            }
                        } catch (IOException e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                        }
                    } else {
                        logger.log(Level.WARNING, "skipped output entry: " + entry.getKey());
                    }
                }
            };
        }

        protected class HbzAlmaCatalogEntityBuilder extends CatalogEntityBuilder implements CatalogEntityRdfOutput {

            protected HbzAlmaCatalogEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                RouteRdfXContentParams routeRdfXContentParams =
                        new RouteRdfXContentParams(catalogEntityWorkerPool.getNamespaceContext(), getIndex(), getType());
                routeRdfXContentParams.setHandler(createHandler());
                if (settings.get("format") != null) {
                    state.getResource().add("format", settings.get("format"));
                }
                if (settings.get("type") != null) {
                    state.getResource().add("type", settings.get("type"));
                }
                if (settings.get("collection") != null) {
                    state.getResource().add("collection", settings.get("collection"));
                }
                state.getResource().newResource("rec")
                        .add("created", catalogEntityWorkerPool.getCreatedFormatted())
                        .add("modified", catalogEntityWorkerPool.getLastModifiedFormatted())
                        .add("indexed", Instant.now().toString());
                try (RdfContentBuilder<RouteRdfXContentParams> builder = RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams)) {
                    builder.receive(state.getResource());
                }
            }

            @Override
            public String getIndex() {
                return "index";
            }

            @Override
            public String getType() {
                return "type";
            }

            @Override
            public RouteRdfXContent.RouteHandler createHandler() {
                return newHandler();
            }
        }
    }

    public class HbzAlmaWorker extends ConverterWorker<String> {

        private final Logger logger = Logger.getLogger(HbzAlmaWorker.class.getName());

        private final HbzAlmaWorkerPool pool;

        public HbzAlmaWorker(Settings settings, HbzAlmaWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            InputStream tarInputStream = ConverterInput.getInputStream(uri, settings);
            Objects.requireNonNull(tarInputStream);
            TarArchiveInputStream tarArchiveInputStream = new TarArchiveInputStream(tarInputStream);
            TarArchiveEntry entry;
            while ((entry = tarArchiveInputStream.getNextEntry()) != null) {
                logger.log(Level.INFO, "processing name = " + entry.getName());
                Path tempFile = Files.createTempFile(entry.getName(), "xml");
                Files.copy(tarArchiveInputStream, tempFile, StandardCopyOption.REPLACE_EXISTING);
                long size = Files.size(tempFile);
                logger.log(Level.INFO, "size = " + size);
                try (InputStream inputStream = Files.newInputStream(tempFile)) {
                    MarcContentHandler marcContentHandler = new MarcContentHandler();
                    marcContentHandler.setMarcRecordListener(pool.catalogEntityWorkerPool);
                    Marc.builder()
                            .setInputStream(inputStream)
                            .setCharset(StandardCharsets.UTF_8)
                            .setContentHandler(marcContentHandler)
                            .build()
                            .xmlReader()
                            .parse();
                } finally {
                    Files.delete(tempFile);
                }
            }
            tarArchiveInputStream.close();
        }
    }
}

