package org.xbib.interlibrary.catalog.builder.ezb;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.URI;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.time.LocalDate;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.RdfConstants;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.internal.DefaultAnonymousResource;
import org.xbib.content.rdf.io.turtle.TurtleContentParams;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.CatalogEntityFileRdfOutput;
import org.xbib.interlibrary.catalog.builder.Converter;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.ConverterWorker;
import org.xbib.interlibrary.catalog.builder.ConverterWorkerPool;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.settings.Settings;
import static org.xbib.content.rdf.RdfContentFactory.turtleBuilder;

/**
 * Fetch EZB web service of german inter library loan codes (ILL).
 */
public class EZBWeb extends Converter<String, EZBWeb.EZBWebWorkerPool> {

    private static final Logger logger = Logger.getLogger(EZBWeb.class.getName());

    public EZBWeb() {
    }

    @Override
    public EZBWebWorkerPool newPool(Settings settings) {
        return new EZBWebWorkerPool(settings);
    }

    public class EZBWebWorkerPool extends ConverterWorkerPool<String> {

        private final Logger logger = Logger.getLogger(EZBWebWorkerPool.class.getName());

        public EZBWebWorkerPool(Settings settings) {
            super("ezb-web", settings);
            String[] beforeCommand = settings.getAsArray("beforeCommand");
            if (beforeCommand != null && beforeCommand.length > 0) {
                executeCommand(beforeCommand);
            }
        }

        @Override
        public void close() throws IOException {
            super.close();
            String[] afterCommand = settings.getAsArray("afterCommand");
            if (afterCommand != null && afterCommand.length > 0) {
                executeCommand(afterCommand);
            }
        }

        @Override
        public EZBWebWorker newWorker(String name, int count) throws IOException {
            return new EZBWebWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        @Override
        protected void openInput() throws IOException, InterruptedException {
            super.openInput();
            //getConverterInput().createFileMap(settings).forEach(getQueue()::add);
            logger.log(Level.INFO, "file map = " + getConverterInput().getFileMap());
            if (!getConverterInput().getFileMap().containsKey("ezb")) {
                logger.log(Level.WARNING, "no 'ezb' input URL found");
                return;
            }
            String baseURL = getConverterInput().getFileMap().get("ezb").getFirst();
            String identifier = settings.get("identifier");
            if (identifier != null) {
                StringBuilder sb = new StringBuilder(baseURL);
                // insert '-' at second last position, which is required for EZB Web API
                sb.append(identifier).insert(sb.length() - 1, '-');
                submit(newWorkerRequest(sb.toString(), baseURL));
            } else {
                if (getConverterInput().getFileMap().containsKey("zdb-id-file")) {
                    String zdbIdFile = getConverterInput().getFileMap().get("zdb-id-file").getFirst();
                    InputStream inputStream = ConverterInput.getInputStream(zdbIdFile, settings);
                    try (BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(inputStream))) {
                        bufferedReader.lines().forEach(line -> {
                            StringBuilder sb = new StringBuilder(baseURL);
                            // insert '-' at second last position, which is required for EZB Web API
                            sb.append(line).insert(sb.length() - 1, '-');
                            submit(newWorkerRequest(sb.toString(), baseURL));
                        });
                    }
                } else {
                    logger.log(Level.WARNING, "no file found of id 'zdb-id-file', please generate one for traversing the EZB Web API by ZDB-ID, see zdb-ezb-file");
                }
            }
            logger.log(Level.INFO, () -> "list of ZDB IDs from file with EZB relations complete");
        }
    }

    private static final String EMPTY = "";

    public class EZBWebWorker extends ConverterWorker<String> {

        private final Logger logger = Logger.getLogger(EZBWebWorker.class.getName());

        private final CatalogEntityFileRdfOutput catalogEntityFileOutput;

        private final IRINamespaceContext namespaceContext;

        private RdfContentBuilder<?> turtleBuilder;

        public EZBWebWorker(Settings settings, EZBWebWorkerPool pool) throws IOException {
            super(settings);
            this.catalogEntityFileOutput = new CatalogEntityFileRdfOutput(pool.getConverterOutput());
            this.namespaceContext = createNamespaceContext();
            if (pool.getConverterOutput().getMap().containsKey("turtle")) {
                OutputStream outputStream = pool.getConverterOutput().getMap().get("turtle").getOut();
                TurtleContentParams turtleParams = new TurtleContentParams(namespaceContext, true);
                turtleBuilder = turtleBuilder(turtleParams, outputStream);
            }
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            int pos = uri.lastIndexOf('/');
            String zdbidWithHyphen = pos > 0 ? uri.substring(pos + 1) : uri;
            // remove hyphen from ZDB-ID
            StringBuilder sb = new StringBuilder().append(zdbidWithHyphen);
            sb.deleteCharAt(sb.length() - 2);
            String zdbid = sb.toString();
            try {
                process(zdbid, getContent(uri));
            } catch (Exception e) {
                logger.log(Level.WARNING, uri + " " + e.getMessage(), e);
            }
        }

        private void process(String zdbid, String body) {
            body.lines().skip(3).forEach(line -> {
                String[] s = line.split("\\t");
                try {
                    int i = 0;
                    String sigel = i < s.length ? s[i++] : EMPTY;
                    String isil = i < s.length ? s[i++] : EMPTY;
                    if (i < s.length) {
                        i++;
                    }
                    String code1 = i < s.length ? s[i++] : EMPTY;
                    String code2 = i < s.length ? s[i++] : EMPTY;
                    String code3 = i < s.length ? s[i++] : EMPTY;
                    String comment = i < s.length ? s[i++] : EMPTY;
                    String firstDate = i < s.length ? s[i++] : EMPTY;
                    String firstVolume = i < s.length ? s[i++] : EMPTY;
                    String firstIssue = i < s.length ? s[i++] : EMPTY;
                    String lastDate = i < s.length ? s[i++] : EMPTY;
                    String lastVolume = i < s.length ? s[i++] : EMPTY;
                    String lastIssue = i < s.length ? s[i++] : EMPTY;
                    String movingWall = i < s.length ? s[i++] : EMPTY;
                    String licenseType = i < s.length ? s[i] : EMPTY;

                    // AAAAA entry ("free serials")
                    if ("AAAAA".equals(sigel)) {
                        isil = "DE-ALL";
                    }
                    String key = zdbid + "." + isil + "."
                            + firstDate + "." + firstVolume + "." + firstIssue + "."
                            + lastDate + "." + lastVolume + "." + lastIssue + "."
                            + (movingWall.isEmpty() ? "0" : movingWall);
                    // fix to null for resource add() method
                    if ("0".equals(firstVolume)) {
                        firstVolume = null;
                    }
                    if ("0".equals(firstIssue)) {
                        firstIssue = null;
                    }
                    if ("0".equals(lastVolume)) {
                        lastVolume = null;
                    }
                    if ("0".equals(lastIssue)) {
                        lastIssue = null;
                    }
                    if (firstDate.isEmpty()) {
                        firstDate = null;
                    }
                    if (lastDate.isEmpty()) {
                        lastDate = null;
                    }
                    if (movingWall.isEmpty()) {
                        movingWall = null;
                    }
                    Resource resource = new DefaultAnonymousResource();
                    resource.add("dc:identifier", key)
                            .add("xbib:identifier", zdbid)
                            .add("xbib:isil", isil)
                            .add("xbib:firstDate", firstDate)
                            .add("xbib:firstVolume", firstVolume)
                            .add("xbib:firstIssue", firstIssue)
                            .add("xbib:lastDate", lastDate)
                            .add("xbib:lastVolume", lastVolume)
                            .add("xbib:lastIssue", lastIssue)
                            .add("xbib:movingWall", movingWall)
                            .add("xbib:interlibraryloanCode",
                                    // replace empty value with "x" (unknown)
                                    (code1.isEmpty() ? "x" : code1)
                                            + (code2.isEmpty() ? "x" : code2)
                                            + (code3.isEmpty() ? "x" : code3))
                            .add("xbib:comment", comment)
                            .add("xbib:licenseType", licenseType)
                            .add("xbib:modified", LocalDate.now().toString());
                    RouteRdfXContentParams params = new RouteRdfXContentParams(namespaceContext,
                            catalogEntityFileOutput.getIndex(), catalogEntityFileOutput.getType());
                    params.setHandler(catalogEntityFileOutput.createHandler());
                    try (RdfContentBuilder<?> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                        builder.receive(resource);
                    }
                    if (turtleBuilder != null) {
                        turtleBuilder.receive(resource);
                    }
                } catch (Exception e) {
                    logger.log(Level.SEVERE, zdbid + ": error: " + e.getMessage(), e);
                }
            });
        }

        private String getContent(String urlString) throws Exception {
            URL url = URI.create(urlString).toURL();
            try (InputStream inputStream = url.openStream()) {
                return new String(inputStream.readAllBytes(), StandardCharsets.UTF_8);
            }
        }
    }

    private void executeCommand(String[] command) {
        ProcessBuilder processBuilder = new ProcessBuilder();
        processBuilder.command(command);
        try {
            Process process = processBuilder.start();
            int exitCode = process.waitFor();
            logger.log(Level.INFO, "command " + Arrays.asList(command) + " exited with " + exitCode);
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    private static IRINamespaceContext createNamespaceContext() {
        IRINamespaceContext namespaceContext = IRINamespaceContext.newInstance();
        namespaceContext.addNamespace(RdfConstants.NS_PREFIX, RdfConstants.NS_URI);
        namespaceContext.addNamespace("dc", "http://purl.org/dc/elements/1.1/");
        namespaceContext.addNamespace("xbib", "http://xbib.org/elements/1.0/");
        return namespaceContext;
    }
}
