package org.xbib.interlibrary.catalog.builder.gnd;

import java.io.IOException;
import java.io.InputStream;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.Resource;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.Triple;
import org.xbib.content.rdf.io.rdfxml.RdfXmlContentParser;
import org.xbib.content.rdf.io.xml.XmlHandler;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.interlibrary.catalog.builder.oai.ConverterOAI;
import org.xbib.interlibrary.catalog.builder.oai.OAICatalogEntityFileRdfOutput;
import org.xbib.oai.xml.SimpleMetadataHandler;
import org.xbib.settings.Settings;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

public class GNDRdfXmlOAI extends ConverterOAI {

    public GNDRdfXmlOAI() {
    }

    @Override
    public GNDRdfXmlOAIWorkerPool newPool(Settings settings) {
        return new GNDRdfXmlOAIWorkerPool(settings);
    }

    public class GNDRdfXmlOAIWorkerPool extends ConverterOAI.ConverterOAIWorkerPool {

        public GNDRdfXmlOAIWorkerPool(Settings settings) {
            super(settings);
        }

        @Override
        public GNDRdfXmlOAIWorker newWorker(String name, int count) {
            return new GNDRdfXmlOAIWorker(settings, this);
        }
    }

    public class GNDRdfXmlOAIWorker extends ConverterOAIWorker {

        private final OAICatalogEntityFileRdfOutput catalogEntityFileOutput;

        public GNDRdfXmlOAIWorker(Settings settings, GNDRdfXmlOAIWorkerPool pool) {
            super(settings, pool);
            this.catalogEntityFileOutput = new OAICatalogEntityFileRdfOutput(pool.getConverterOutput());
        }

        @Override
        protected SimpleMetadataHandler newMetadataHandler() throws IOException {
            RouteRdfXContentParams params = new RouteRdfXContentParams(createNamespaceContext(),
                    catalogEntityFileOutput.getIndex(), catalogEntityFileOutput.getType());
            RdfXmlContentParser<RouteRdfXContentParams> parser = new GNDRdfXmlContentParser(params, null);
            parser.setRdfContentBuilderProvider(() -> RdfXContentFactory.routeRdfXContentBuilder(params));
            return new MyHandler(params, parser.getHandler());
        }

        class MyHandler extends SimpleMetadataHandler {

            final XmlHandler<RouteRdfXContentParams> handler;

            final RouteRdfXContentParams params;

            MyHandler(RouteRdfXContentParams params, XmlHandler<RouteRdfXContentParams> handler) {
                this.params = params;
                this.handler = handler;
            }

            @Override
            public void startDocument() throws SAXException {
                handler.startDocument();
            }

            @Override
            public void endDocument() throws SAXException {
                params.setHandler(catalogEntityFileOutput.createHandler());
                handler.endDocument();
            }

            @Override
            public void startPrefixMapping(String string, String string1) throws SAXException {
                handler.startPrefixMapping(string, string1);
            }

            @Override
            public void endPrefixMapping(String string) throws SAXException {
                handler.endPrefixMapping(string);
            }

            @Override
            public void startElement(String ns, String localname, String string2, Attributes atrbts) throws SAXException {
                handler.startElement(ns, localname, string2, atrbts);
            }

            @Override
            public void endElement(String ns, String localname, String string2) throws SAXException {
                handler.endElement(ns, localname, string2);
            }

            @Override
            public void characters(char[] chars, int i, int i1) throws SAXException {
                handler.characters(chars, i, i1);
            }
        }

        private class GNDRdfXmlContentParser extends RdfXmlContentParser<RouteRdfXContentParams> {

            private Resource lastSubject;

            private String id;

            private final RouteRdfXContentParams params;

            private GNDRdfXmlContentParser(RouteRdfXContentParams params, InputStream inputStream) {
                super(params, inputStream);
                this.params = params;
            }

            @Override
            public RdfXmlContentParser<RouteRdfXContentParams> parse(XMLReader reader, InputSource source)
                    throws IOException, SAXException {
                super.parse(reader, source);
                RdfContentBuilder<RouteRdfXContentParams> builder = getBuilder();
                if (builder != null && lastSubject != null) {
                    builder.endStream();
                    catalogEntityFileOutput.write(id, params.getGenerator().get());
                    builder.close();
                }
                return this;
            }

            @Override
            protected void yieldTriple(Triple t) throws IOException {
                RdfContentBuilder<RouteRdfXContentParams> builder = getBuilder();
                if (lastSubject == null || !t.subject().equals(lastSubject)) {
                    if (builder != null) {
                        builder.endStream();
                        catalogEntityFileOutput.write(id, params.getGenerator().get());
                        builder.close();
                    }
                    builder = RdfXContentFactory.routeRdfXContentBuilder(params);
                    builder.startStream();
                    builder.receive(t.subject().id());
                    lastSubject = t.subject();
                }
                if ("http://d-nb.info/standards/elementset/gnd#gndIdentifier".equals(t.predicate().toString())) {
                    id = t.object().toString();
                }
                if (builder != null) {
                    builder.receive(t);
                }
            }
        }
    }

    private static IRINamespaceContext createNamespaceContext() {
        IRINamespaceContext namespaceContext = IRINamespaceContext.newInstance();
        namespaceContext.addNamespace("dc", "http://purl.org/dc/elements/1.1/");
        namespaceContext.addNamespace("geo", "http://rdvocab.info/");
        namespaceContext.addNamespace("rda", "http://purl.org/dc/elements/1.1/");
        namespaceContext.addNamespace("foaf", "http://xmlns.com/foaf/0.1/");
        namespaceContext.addNamespace("sf", "http://www.opengis.net/ont/sf#");
        namespaceContext.addNamespace("isbd", "http://iflastandards.info/ns/isbd/elements/");
        namespaceContext.addNamespace("gnd", "http://d-nb.info/standards/elementset/gnd#");
        namespaceContext.addNamespace("dcterms", "http://purl.org/dc/terms/");
        namespaceContext.addNamespace("rdfs", "http://www.w3.org/2000/01/rdf-schema#");
        namespaceContext.addNamespace("marcRole", "http://id.loc.gov/vocabulary/relators/");
        namespaceContext.addNamespace("lib", "http://purl.org/library/");
        namespaceContext.addNamespace("umbel", "http://umbel.org/umbel#");
        namespaceContext.addNamespace("bibo", "http://purl.org/ontology/bibo/");
        namespaceContext.addNamespace("owl", "http://www.w3.org/2002/07/owl#");
        namespaceContext.addNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
        namespaceContext.addNamespace("skos", "http://www.w3.org/2004/02/skos/core#");
        namespaceContext.addNamespace("geosparql", "http://www.opengis.net/ont/geosparql#");
        return namespaceContext;
    }
}
