package org.xbib.interlibrary.catalog.builder.marc;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.Normalizer;
import java.time.Instant;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContent;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.CatalogEntityRdfOutput;
import org.xbib.interlibrary.catalog.builder.Converter;
import org.xbib.interlibrary.catalog.builder.ConverterInput;
import org.xbib.interlibrary.catalog.builder.ConverterOutput;
import org.xbib.interlibrary.catalog.builder.ConverterWorker;
import org.xbib.interlibrary.catalog.builder.ConverterWorkerPool;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilder;
import org.xbib.interlibrary.catalog.entities.CatalogEntityBuilderState;
import org.xbib.interlibrary.catalog.entities.CatalogEntityWorkerPool;
import org.xbib.interlibrary.catalog.entities.ISILTransformer;
import org.xbib.marc.Marc;
import org.xbib.marc.transformer.value.MarcValueTransformers;
import org.xbib.settings.Settings;

/**
 * Tool for MARC bibliographic data.
 */
public class MarcBibliographic extends Converter<String, MarcBibliographic.MarcBibliographicWorkerPool> {

    public MarcBibliographic() {
    }

    @Override
    public MarcBibliographicWorkerPool newPool(Settings settings) throws IOException {
        return new MarcBibliographicWorkerPool(settings);
    }

    public class MarcBibliographicWorkerPool extends ConverterWorkerPool<String> {

        private final IRINamespaceContext namespaceContext;

        protected ISILTransformer isilTransformer;

        public MarcBibliographicWorkerPool(Settings settings) throws IOException {
            super("marc-bib", settings);
            this.namespaceContext = IRINamespaceContext.newInstance(MarcBibliographic.class);
            this.isilTransformer = new ISILTransformer(settings);
        }

        @Override
        public MarcBibliographicWorker newWorker(String name, int count) {
            return new MarcBibliographicWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }
    }


    public class MarcBibliographicWorker extends ConverterWorker<String> {

        private final MarcBibliographicWorkerPool pool;

        public MarcBibliographicWorker(Settings settings, MarcBibliographicWorkerPool pool) {
            super(settings);
            this.pool = pool;
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            MarcValueTransformers marcValueTransformers = newMarcValueTransformers();
            try (InputStream inputStream = ConverterInput.getInputStream(uri, settings);
                 CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool =
                         new CatalogEntityWorkerPool<>(settings, new DefaultWorkerPoolMarcRecordListener(),
                                 MarcBibliographicCatalogEntityBuilder::new, pool)) {
                doExecute(inputStream, catalogEntityWorkerPool, marcValueTransformers);
            }
        }

        protected MarcValueTransformers newMarcValueTransformers() {
            MarcValueTransformers marcValueTransformers = new MarcValueTransformers();
            marcValueTransformers.setMarcValueTransformer(value -> Normalizer.normalize(value, Normalizer.Form.NFC));
            return marcValueTransformers;
        }

        protected void doExecute(InputStream inputStream,
                                 CatalogEntityWorkerPool<CatalogEntityBuilder> catalogEntityWorkerPool,
                                 MarcValueTransformers marcValueTransformers) throws IOException {
            String dialect = settings.get("dialect");
            String charset = settings.get("charset", "UTF-8");
            if ("bibliomondo".equals(dialect)) {
                Marc marc = Marc.builder()
                        .setInputStream(inputStream)
                        .setMarcListener(catalogEntityWorkerPool)
                        .setMarcValueTransformers(marcValueTransformers)
                        .setCharset(Charset.forName(charset))
                        .build();
                marc.wrapIntoCollection(marc.bibliomondo());
            } else {
                Marc.builder()
                        .setInputStream(inputStream)
                        .setMarcRecordListener(catalogEntityWorkerPool)
                        .setMarcValueTransformers(marcValueTransformers)
                        .setCharset(Charset.forName(charset))
                        .build()
                        .writeRecords();
            }
        }

        protected class MarcBibliographicCatalogEntityBuilder
                extends CatalogEntityBuilder implements CatalogEntityRdfOutput {

            private final Logger logger = Logger.getLogger(MarcBibliographicCatalogEntityBuilder.class.getName());

            protected MarcBibliographicCatalogEntityBuilder(Settings settings, AtomicLong counter) throws IOException {
                super(settings, counter);
            }

            @Override
            public void afterFinishState(CatalogEntityBuilderState state) throws IOException {
                RouteRdfXContentParams routeRdfXContentParams =
                        new RouteRdfXContentParams(pool.namespaceContext, getIndex(), getType());
                routeRdfXContentParams.setHandler(createHandler());
                if (settings.get("format") != null) {
                    state.getResource().add("format", settings.get("format"));
                }
                if (settings.get("type") != null) {
                    state.getResource().add("type", settings.get("type"));
                }
                if (settings.get("collection") != null) {
                    state.getResource().add("collection", settings.get("collection"));
                }
                state.getResource().newResource("rec").add("modified", Instant.now().toString());
                try (RdfContentBuilder<RouteRdfXContentParams> builder =
                        RdfXContentFactory.routeRdfXContentBuilder(routeRdfXContentParams)) {
                    builder.receive(state.getResource());
                }
            }

            @Override
            public String getIndex() {
                return "index";
            }

            @Override
            public String getType() {
                return "type";
            }

            @Override
            public RouteRdfXContent.RouteHandler createHandler() {
                return (content, p) -> {
                    for (Map.Entry<String, ConverterOutput.Entry> entry : pool.getConverterOutput().getMap().entrySet()) {
                        if (entry.getKey().startsWith("file")) {
                            try {
                                synchronized (pool) {
                                    entry.getValue().getOut().write(content.getBytes(StandardCharsets.UTF_8));
                                    entry.getValue().getOut().write('\n');
                                    entry.getValue().getOut().flush();
                                }
                            } catch (IOException e) {
                                logger.log(Level.SEVERE, e.getMessage(), e);
                            }
                        }
                    }
                };
            }
        }
    }
}
