package org.xbib.interlibrary.catalog.builder.oai;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.text.ParseException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.stream.XMLStreamException;
import org.xbib.content.rdf.RdfContentBuilder;
import org.xbib.content.rdf.RdfXContentFactory;
import org.xbib.content.rdf.RouteRdfXContentParams;
import org.xbib.content.rdf.io.ntriple.NTripleContentParams;
import org.xbib.content.rdf.io.turtle.TurtleContentParams;
import org.xbib.content.resource.IRINamespaceContext;
import org.xbib.interlibrary.api.action.worker.WorkerRequest;
import org.xbib.interlibrary.catalog.builder.Converter;
import org.xbib.interlibrary.catalog.builder.ConverterOutput;
import org.xbib.interlibrary.catalog.builder.ConverterWorker;
import org.xbib.interlibrary.catalog.builder.ConverterWorkerPool;
import org.xbib.interlibrary.catalog.builder.StringWorkerRequest;
import org.xbib.oai.OAIExceptionHandler;
import org.xbib.oai.client.OAIClient;
import org.xbib.oai.client.identify.IdentifyResponse;
import org.xbib.oai.exceptions.OAIException;
import org.xbib.oai.rdf.RdfResourceHandler;
import org.xbib.oai.rdf.RdfSimpleMetadataHandler;
import org.xbib.oai.xml.MetadataHandler;
import org.xbib.oai.xml.SimpleMetadataHandler;
import org.xbib.oai.xml.XmlSimpleMetadataHandler;
import org.xbib.settings.Settings;
import org.xbib.time.chronic.Chronic;
import org.xbib.time.chronic.Span;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import static org.xbib.content.rdf.RdfContentFactory.ntripleBuilder;
import static org.xbib.content.rdf.RdfContentFactory.turtleBuilder;

public class ConverterOAI extends Converter<String, ConverterOAI.ConverterOAIWorkerPool> {

    public ConverterOAI() {
    }

    @Override
    public ConverterOAIWorkerPool newPool(Settings settings) {
        return new ConverterOAIWorkerPool(settings);
    }

    public class ConverterOAIWorkerPool extends ConverterWorkerPool<String> {

        private final Settings oaiSettings;

        private final IRINamespaceContext namespaceContext;

        private final AtomicLong errorCounter;

        public ConverterOAIWorkerPool(Settings settings) {
            super("oai", settings);
            this.oaiSettings = settings.getGroups("input").get("oai");
            this.namespaceContext = IRINamespaceContext.newInstance(ConverterOAI.this.getClass());
            this.errorCounter = new AtomicLong(0);
        }

        @Override
        public ConverterOAIWorker newWorker(String name, int count) throws IOException {
            return new ConverterOAIWorker(settings, this);
        }

        @Override
        public WorkerRequest<String> newWorkerRequest(Object input, String origin) {
            return input != null ? new StringWorkerRequest(input.toString(), origin) : null;
        }

        @Override
        protected void openInput() throws IOException, InterruptedException {
            submit(newWorkerRequest(oaiSettings.get("uri"), "oai"));
            super.openInput();
        }

        public long getErrorCounter() {
            return errorCounter.get();
        }
    }

    public class ConverterOAIWorker extends ConverterWorker<String> {

        private final Logger logger = Logger.getLogger(ConverterOAIWorker.class.getName());

        private final OAICatalogEntityFileRdfOutput catalogEntityFileOutput;

        private final Settings oaiSettings;

        protected final ConverterOAIWorkerPool pool;

        public ConverterOAIWorker(Settings settings, ConverterOAIWorkerPool pool) {
            super(settings);
            this.pool = pool;
            this.oaiSettings = settings.getGroups("input").get("oai");
            this.catalogEntityFileOutput = new OAICatalogEntityFileRdfOutput(pool.getConverterOutput());
        }

        @Override
        protected void doExecute(String uri) throws IOException {
            String metadataPrefix = oaiSettings.get("metadataPrefix");
            String set = oaiSettings.get("set");
            String granularity = oaiSettings.get("granularity");
            Instant from = fromChrono(oaiSettings.get("from"));
            Instant until = fromChrono(oaiSettings.get("until"));
            String baseStr = oaiSettings.get("base");
            Instant base = Instant.parse(baseStr != null ? baseStr : "2010-01-01T00:00:00Z");
            OAIClient oaiClient = new OAIClient(uri);
            oaiClient.setUserAgent("hbz OAI client");
            if (granularity == null) {
                try {
                    IdentifyResponse identifyResponse = oaiClient.identify();
                    granularity = identifyResponse.getGranularity();
                } catch (OAIException e) {
                    logger.log(Level.WARNING, e.getMessage(), e);
                } catch (InterruptedException e) {
                    logger.log(Level.WARNING, "interrupted");
                }
            }
            if (granularity == null) {
                granularity = "YYYY-MM-DD";
            }
            logger.log(Level.INFO, "using granularity = " + granularity);
            DateTimeFormatter dateTimeFormatter = switch (granularity) {
                case "yyyy-MM-dd'T'HH:mm:ss'Z'", "yyyy-MM-ddTHH:mm:ssZ" ->
                        DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ss'Z'").withZone(ZoneId.of("UTC"));
                default -> DateTimeFormatter.ofPattern("yyyy-MM-dd").withZone(ZoneId.of("UTC"));
            };
            MetadataHandler handler = getMetadataHandler();
            OAIExceptionHandler exceptionHandler = getExceptionHandler();
            try {
                if (handler != null) {
                    oaiClient.listRecords(metadataPrefix, set,
                            dateTimeFormatter, from, until, base, handler, exceptionHandler, null);
                } else {
                    oaiClient.listRecords(metadataPrefix, set,
                            dateTimeFormatter, from, until, base, null, exceptionHandler, this::process);
                }
            } catch (OAIException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                pool.errorCounter.incrementAndGet();
                throw new IllegalStateException(e);
            }
        }

        protected void process(InputStream inputStream) {
            for (Map.Entry<String, ConverterOutput.Entry> entry : pool.getConverterOutput().getMap().entrySet()) {
                try {
                    if (entry.getKey().startsWith("file")) {
                        entry.getValue().copy(inputStream);
                    }
                } catch (IOException e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                }
            }
        }

        protected MetadataHandler getMetadataHandler() {
            return null; // override if required
        }

        protected OAIExceptionHandler getExceptionHandler() {
            return null;
        }

        protected SimpleMetadataHandler newMetadataHandler() throws IOException {
            if (settings.containsSetting("handler")) {
                switch (settings.get("handler")) {
                    case "xml" -> {
                        return xmlMetadataHandler();
                    }
                    case "turtle" -> {
                        return turtleMetadataHandler();
                    }
                    case "ntriples" -> {
                        return ntripleMetadataHandler();
                    }
                }
            }
            return new OAISimpleMetadataHandler(this);
        }

        protected String convert(String id, String content) throws IOException {
            return content;
        }

        private Instant fromChrono(String chronoString) {
            try {
                Span span = Chronic.parse(chronoString);
                logger.log(Level.INFO, "found instant from chrono: " + DateTimeFormatter.ISO_INSTANT.format(span.getBeginCalendar()));
                return span.getBeginCalendar().toInstant();
            } catch (ParseException e) {
                logger.log(Level.WARNING, "no chrono found in " + chronoString);
                return Instant.parse(chronoString);
            }
        }

        private SimpleMetadataHandler xmlMetadataHandler() throws IOException {
            try {
                return new XmlPacketHandlerSimple().setWriter(new StringWriter());
            } catch (XMLStreamException e) {
                throw new IOException(e);
            }
        }

        private SimpleMetadataHandler turtleMetadataHandler() throws IOException {
            final RdfSimpleMetadataHandler metadataHandler = new RdfSimpleMetadataHandler();
            final RdfResourceHandler resourceHandler = rdfResourceHandler();
            final TurtleContentParams params = new TurtleContentParams(pool.namespaceContext, false);
            metadataHandler.setHandler(resourceHandler)
                    .setBuilder(turtleBuilder(params));
            return metadataHandler;
        }

        private SimpleMetadataHandler ntripleMetadataHandler() throws IOException {
            final RdfSimpleMetadataHandler metadataHandler = new RdfSimpleMetadataHandler();
            final RdfResourceHandler resourceHandler = rdfResourceHandler();
            final NTripleContentParams params = new NTripleContentParams(pool.namespaceContext);
            metadataHandler.setHandler(resourceHandler)
                    .setBuilder(ntripleBuilder(params));
            return metadataHandler;
        }

        private RdfResourceHandler rdfResourceHandler() {
            NTripleContentParams params = new NTripleContentParams(pool.namespaceContext);
            return new RdfResourceHandler(params);
        }

        private class XmlPacketHandlerSimple extends XmlSimpleMetadataHandler {

            @Override
            public void endDocument() throws SAXException {
                super.endDocument();
                logger.log(Level.FINE, "got XML document " + getIdentifier());
                try {
                    setWriter(new StringWriter());
                } catch (XMLStreamException e) {
                    throw new SAXException(e);
                }
            }
        }

        /**
         *
         */
        public class OAISimpleMetadataHandler extends SimpleMetadataHandler {

            protected final ConverterOAIWorker worker;

            protected final IRINamespaceContext namespaceContext;

            private RdfResourceHandler handler;

            public OAISimpleMetadataHandler(ConverterOAIWorker worker) {
                this.worker = worker;
                namespaceContext = IRINamespaceContext.newInstance();
                namespaceContext.addNamespace("", "http://www.openarchives.org/OAI/2.0/oai_dc/");
                namespaceContext.addNamespace("dc", "http://purl.org/dc/elements/1.1/");
            }

            @Override
            public void startDocument() throws SAXException {
                this.handler = worker.rdfResourceHandler();
                handler.setDefaultNamespace("", "http://www.openarchives.org/OAI/2.0/oai_dc/");
                handler.startDocument();
            }

            @Override
            public void endDocument() throws SAXException {
                handler.endDocument();
                try {
                    RouteRdfXContentParams params = new RouteRdfXContentParams(namespaceContext,
                            catalogEntityFileOutput.getIndex(), catalogEntityFileOutput.getType());
                    params.setHandler((content, p) -> {
                        // custom convert OAI content
                        String data = convert(getHeader().getIdentifier(), content);
                        // pass it to file output handler
                        catalogEntityFileOutput.createHandler(getHeader().getIdentifier())
                                .complete(data, p);
                    });
                    try (RdfContentBuilder<?> builder = RdfXContentFactory.routeRdfXContentBuilder(params)) {
                        builder.receive(handler.getResource());
                    }
                } catch (IOException e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                    throw new SAXException(e);
                }
            }

            @Override
            public void startPrefixMapping(String string, String string1) throws SAXException {
                handler.startPrefixMapping(string, string1);
            }

            @Override
            public void endPrefixMapping(String string) throws SAXException {
                handler.endPrefixMapping(string);
            }

            @Override
            public void startElement(String ns, String localname, String string2, Attributes atrbts) throws SAXException {
                handler.startElement(ns, localname, string2, atrbts);
            }

            @Override
            public void endElement(String ns, String localname, String string2) throws SAXException {
                handler.endElement(ns, localname, string2);
            }

            @Override
            public void characters(char[] chars, int i, int i1) throws SAXException {
                handler.characters(chars, i, i1);
            }
        }
    }
}
